#!/usr/bin/perl -w
# $Id: ace2xml.pl 256 2013-04-10 19:02:56Z pjohnson $

# ace2xml.pl -- Takes ACE format (from phred/phrap/consed) + quality
# scores in FASTA format and converts to XML conforming to piim_r.dtd
#
# Created 2 May 2007 by Philip Johnson
# Updated 15 Jul 2008 to conform to new XML format that maintains linkage

use strict;
use Getopt::Std;
use lib <~>.'/bin';
use FF_Index 3;

{
    my %opts;
    $opts{o} = '';
    if (!getopts("i:q:o:c:", \%opts)  ||  !defined $opts{i}  ||
        !defined $opts{q}) {
        print "usage: $0 -i <input.ace> -q <input.qual> [-o <output-file-prefix>] [-c <contig>]

  Converts ACE (phrap)-formated contig assemblies into XML required
  for input to PIIM.

  The <input.qual> file should contain fasta-esque formatting of
  quality scores for each read used in the ACE file.  XML for each
  contig will be placed in a separate output file named
  '<outfile-file-prefix><contig-name>.xml'.  If the -c option is
  specified, only that contig will be parsed.

  Reads are assumed to be paired (from the same chromosome) if their
  identifiers are identical up to a non-word character
  (i.e. APZO11644.b2 and APZO11644.g2).\n\n";
        exit 1;
    }


    my $qualIdx = FF_Index->new($opts{q}, \&FindNextQualEntry, inMemory=>1);

    open ACEFILE, $opts{i} or  die "Error opening ACE file: '$opts{i}' ($!)\n";
    my ($numContigs) = <ACEFILE> =~ /^AS\s+(\d+)\s/;
    #die "$0 can only handle 1 contig per ace file\n" if ($numContigs > 1);
    
    for (my $i = 0;  $i < $numContigs;  ++$i) {
        while (defined($_ = <ACEFILE>)  &&  !/^CO/) {}
        if ($opts{c}  &&  !/^CO $opts{c}/) {
            next;
        }
        ProcessContig(\*ACEFILE, $qualIdx, $opts{o});
    }

    close ACEFILE;
}

# -----------------------------------------------------------------------------
# Takes ACEFILE queued to start of contig, along with QUALFILE & index
#
sub ProcessContig {
    my ($ACEFILE, $qualIdx, $outfilePre) = @_;

    my ($ctgName, $numBases, $numReads) = /^CO\s+(\S+)\s+(\d+)\s+(\d+)/;
    print STDERR 'Creating ', $outfilePre.$ctgName.'.xml', "..\n";

    while (defined($_ = <$ACEFILE>)  &&  !/^AF/) {} #skip consensus

    my @assm;
    my $totNumReads = 0;
    do {
        ++$totNumReads;
        my ($acc, $ori, $pos) = /^AF\s+(\S+)\s+(U|C)\s+(-?\d+)/;
        if ($acc !~ /_/) { #eliminate non-random reads (primer walks & 454)
            #switch to 0-based coordinates to maintain sanity
            push @assm, [$acc, $ori eq 'C', $pos-1];
        }
    } while (defined($_ = <$ACEFILE>)  &&  /^AF/);
    # overshot, so rewind one line in case important later
    seek($ACEFILE, -length($_), 1);
    @assm = sort {$a->[0] cmp $b->[0]  ||  $a->[2] <=> $b->[2]} @assm;

    # create read index by parsing the RD records
    my %readIdx;
    for (my $i = 0;  $i < $totNumReads;  ++$i) {
        # skip to start of RD record
        while (defined($_ = <$ACEFILE>)  &&  !/^RD/) {}
        my ($acc) = /^RD\s+(\S+)/;
        $readIdx{$acc} = tell($ACEFILE) - length($_);
    }

    #main loop over ctg
    open OUTFILE, '>'.$outfilePre.$ctgName.'.xml';
    print OUTFILE '<?xml version="1.0"?>', "\n";
    print OUTFILE '<!DOCTYPE Contig PUBLIC "" "piim_r.dtd">', "\n";
    print OUTFILE "<Contig>\n";
    for (my $readI = 0;  $readI < @assm;  ++$readI) {
        print OUTFILE "  <Chr>\n";

        if (!exists($readIdx{$assm[$readI][0]})) {
            die "Error: couldn't find read data for $assm[$readI][0]\n";
        }
        seek $ACEFILE, $readIdx{$assm[$readI][0]}, 0;
        ProcessRead(\*OUTFILE, $ACEFILE, $assm[$readI], $qualIdx);

        if ($readI+1 < @assm  &&
            ($assm[$readI][0] =~ /^(\w+)/)[0] eq
            ($assm[$readI+1][0] =~ /^(\w+)/)[0]) {
            #print $assm[$readI][0], "\t", $assm[$readI+1][0], "\n";
            ++$readI;
            seek $ACEFILE, $readIdx{$assm[$readI][0]}, 0;
            ProcessRead(\*OUTFILE, $ACEFILE, $assm[$readI], $qualIdx);
        }

        print OUTFILE "  </Chr>\n";
    }
    print OUTFILE "</Contig>\n";
    close OUTFILE;
}


# -----------------------------------------------------------------------------
# Callback for qual file indexing by accession
#
sub FindNextQualEntry {
    my ($fh, $record) = @_;

    while (defined($_ = <$fh>)  &&  !/^>/) {} #skip to next
    if (defined($_)) {
        ($record->{key}) = /^>(\S+)/;
        $record->{pos} = tell($fh);
        return 1;
    } else {
        return 0;
    }
}

# -----------------------------------------------------------------------------
# Load and concatenate lines (presumably nucleotides) until hitting a
# blank line
#
sub ReadBases {
    my ($ACEFILE) = @_;

    my $seq = [];
    while (defined($_ = <$ACEFILE>)  &&  !/^$/) {
        chomp;
        push @$seq, split(//, uc $_);
    }

    return $seq;
}

# -----------------------------------------------------------------------------
# Read and concatenate q-scores for a single read
#
sub ReadQuals {
    my ($QUALFILE) = @_;

    return undef if !defined($QUALFILE);

    my $quals = [];
    while (defined($_ = <$QUALFILE>)  &&  !/^>/) {
        chomp;
        push @$quals, split(/ +/, $_);
    }

    return $quals;
}

# -----------------------------------------------------------------------------
# Load trimming information
#
sub ReadQualTrimming {
    my ($ACEFILE) = @_;

    while (defined($_ = <$ACEFILE>)  &&  !/^QA/  &&  !/^RD/) {}
    if (/^QA/) {
        my (undef, undef, undef, $left, $right) = split;
        return ($left-1, $right-1);
    } else {
        return (0,0);
    }
}

# -----------------------------------------------------------------------------
# Load read (bases & q-scores) that we're queued up for reading
sub ProcessRead {
    my ($OUTFILE, $ACEFILE, $assmInfo, $qualIdx) = @_;

    $_ = <$ACEFILE>;
    my ($acc, $paddedLength) = /^RD\s+(\S+)\s+(\d+)/;

    die "Error: order mismatch between assembly (AF:$assmInfo->[0]) and reads (RD:$acc)\n" if $acc ne $assmInfo->[0];

    my $bases = ReadBases($ACEFILE); #already complemented if necessary
    my $quals = ReadQuals($qualIdx->{$acc}) or
        die "ERROR: no quals for '$acc'\n";

    if ($assmInfo->[1]) { #complement -- so flip quals
        @$quals = reverse(@$quals);
    }

    for (my $i = 0;  $i < @$bases;  ++$i) { #insert gaps into quals
        if ($bases->[$i] eq '*') {
            splice @$quals, $i, 0, 0;
            $bases->[$i] = '-';
        }
    }

    my ($trimL, $trimR) = ReadQualTrimming($ACEFILE);
    die "Unhandled case: $acc contains no aligned bases.\n" if ($trimL < 0  ||  $trimR < 0);
    splice @$bases, 0, $trimL;
    splice @$bases, $trimR+1-$trimL;
    splice @$quals, 0, $trimL;
    splice @$quals, $trimR+1-$trimL;
    $assmInfo->[2] += $trimL;

    die "Error: base call & q-score mismatch for $acc" if scalar(@$bases) != scalar(@$quals);

    print $OUTFILE '    <Read start="'.$assmInfo->[2]."\" acc=\"$acc\">\n";
    print $OUTFILE '      <Bases>', join('', @$bases), "</Bases>\n";
    print $OUTFILE '      <Quals>', join(' ', @$quals), "</Quals>\n";
    print $OUTFILE "    </Read>\n";
}
