/* $Id: util.hpp 358 2020-03-16 23:03:27Z pjohnson $
  ---------------------------------------------------------------------------
  Useful functions & type definitions
  ---------------------------------------------------------------------------
  Copyright (C) 2009-2011 Philip Johnson

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------
*/
#ifndef UTIL_HPP
#define UTIL_HPP

#include <iostream>
#include <sstream>
#include <stdexcept>
#include <map>
#include <vector>
#include <cmath>
#include <cstring>
#include <typeinfo>
#include <cstdio>
#include <limits>
#include <getopt.h>

using namespace std;

/*---------------------------------------------------------------------------*/
// PRE : string representation
// POST: TTo
template <typename TTo>
inline TTo StringTo(const string &s) {
    istringstream iss(s);
    TTo i; iss >> i;
    if (!iss  ||  !iss.eof()) {
        throw runtime_error("Error converting '"+s+"' to "+typeid(TTo).name());
    }
    return i;
}

/*---------------------------------------------------------------------------*/
// PRE : TFrom
// POST: string representation
template <typename TFrom>
inline const string ToString(const TFrom &i) {
    ostringstream oss;
    oss << i;
    return oss.str();
}

//---------------------------------------------------------------------------

template <typename TContents, typename TIndex>
class CCache {
public:
    typedef CCache<TIndex, TContents> TThisType;
    inline const TContents& operator[] (const TIndex& index) const {
        typename TData::const_iterator it = m_Data.find(index);
        if (it == m_Data.end()) {
            ostringstream oss;
            oss << index;
            throw runtime_error("missed cache: " + oss.str());
        } else {
            return it->second;
        }
    }
    inline const TContents* Get(const TIndex& index) const {
        typename TData::const_iterator it = m_Data.find(index);
        if (it == m_Data.end()) {
            return NULL;
        } else {
            return &(it->second);
        }
    }
    inline TContents& Set(const TIndex& i) {
        return m_Data[i];
    }

    inline void Reset(void) {
        m_Data.clear();
    }
    inline void clear(void) {
        m_Data.clear();
    }
    inline size_t size(void) const {
        return m_Data.size();
    }


private:
    typedef map< TIndex, TContents > TData;
    TData m_Data;
};

template <typename TContents, typename TIndex = int> class CCache;

template <typename TContents>
class CCache<TContents, int> {
public:
    typedef CCache<TContents> TThisType;
    inline const TContents& operator[] (unsigned int i) const {
        if (i >= m_List.size()  ||  !m_List[i].second) {
            ostringstream oss;
            oss << i;
            throw runtime_error("missed cache: " + oss.str());
        } else {
            return m_List[i].first;
        }
    }

    inline TContents& Set(unsigned int i) {
        if (i >= m_List.size()) {
            m_List.resize(i+1);
        }
        m_List[i].second = true;
        return m_List[i].first;
    }

    inline void Reset(void) {
        m_List.clear();
    }
    inline void clear(void) {
        m_List.clear();
    }

private:
    typedef vector< pair<TContents,bool> > TList;
    TContents m_Null;
    TList m_List;
};

/*---------------------------------------------------------------------------*/

class COpt {
public:
    enum EHasArg {
        eNoArg = 0,
        eReqArg,
        eOptArg
    };

public:
    COpt(void) : m_IsSet(false), m_Required(false), m_OverrideRequired(false),
                 m_HasArg(eReqArg), m_AllowMultiple(false) {}
    void operator= (const char* data) { m_Data = data; m_IsSet = true; }
    void operator+= (const char* data) {
        if (m_IsSet) { m_Data += string(" ") + data; } else { (*this) = data;}}
    COpt& SetRequired(bool b) { m_Required = b; return *this; }
    COpt& SetOverrideRequired(bool b) { m_OverrideRequired = b; return *this; }
    COpt& HasArg(EHasArg a) { m_HasArg = a; return *this;}
    COpt& SetFlag(void) { m_HasArg = eNoArg; return *this;}
    COpt& SetDesc(const string& d) { m_Descript = d; return *this;}
    COpt& SetAllowMultiple(void) { m_AllowMultiple = true; return *this;}

    bool IsRequired(void) const { return m_Required; }
    bool IsOverrideRequired(void) const { return m_OverrideRequired; }
    EHasArg HasArg(void) const { return m_HasArg; }
    const string& GetDesc(void) const { return m_Descript; }
    bool AllowsMultiple(void) const { return m_AllowMultiple; }
    
    //OK? commenting out below line on 8/20/18 to worked around compiler error in g++ 7.3.0 in empirical_damage.cpp
    //operator const char*(void) const { return m_IsSet ? m_Data.c_str() : NULL;}
    operator bool (void) const {return m_IsSet;}
    operator string(void) const { return m_Data; }
    bool operator== (const char *s) const { return m_Data == s; }
    bool operator!= (const char *s) const { return m_Data != s; }
    friend string operator+ (const COpt &o, const char *s) {
        return o.m_Data + s; }
    friend string operator+ (const char *s, const COpt &o) {
        return s + o.m_Data; }

    string AsString(void) const { return m_Data; }
    double AsDouble(void) const {
        return (AsString() == "inf") ?
            numeric_limits<double>::infinity() : StringTo<double>(AsString());}
    int AsInt(void) const { return StringTo<int>(AsString()); }
    void AsMatrix(unsigned int nrow, unsigned int ncol, double* m) const {
        istringstream iss(m_Data);
        for (unsigned int i = 0;  i < nrow;  ++i) {
            for (unsigned int j = 0;  j < ncol;  ++j) {
                if (iss.eof()) {
                    throw runtime_error("Error converting matrix argument (perhaps insufficient entries?)");
                }
                iss >> m[i*ncol + j];
                if (iss.fail()) {
                    throw runtime_error("Error converting matrix argument (perhaps text where there should be a number?)");
                }
            }
        }
    }

private:
    bool m_IsSet;
    string m_Data;
    bool m_Required;
    bool m_OverrideRequired;
    EHasArg m_HasArg;
    string m_Descript;
    bool m_AllowMultiple;
};

class TOpts : public map<string, COpt> {
    typedef map<string, COpt> TParent;
public:
    COpt& operator[] (const char c) {
        return TParent::operator[](string(1,c));
    }
    COpt& operator[] (const char* c) {
        return TParent::operator[](string(c));
    }
    const COpt& operator[] (const char c) const {
        const_iterator i = find(string(1,c));
        if (i == end()) {
            throw logic_error("Invalid key to TOpts: " + string(1,c));
        }
        return i->second;
    }
    const COpt& operator[] (const char* c) const {
        const_iterator i = find(string(c));
        if (i == end()) {
            throw logic_error("Invalid key to TOpts: " + string(c));
        }
        return i->second;
    }
};

/*---------------------------------------------------------------------------*/

inline void x_ProcessOptions(int argc, char **argv, TOpts &opts,
                              const string& descript)
{
    TOpts defaultOpts(opts);
    string errString, warnString;

    for (int argI = 1;  argI < argc;  ++argI) {
        if (argv[argI][0] == '-') {
            TOpts::iterator optI = opts.find(argv[argI]+1);
            if (optI == opts.end()) { //unknown option
                errString += "\t* unknown option '" + string(argv[argI])+"'\n";
            } else if (optI->second.HasArg() == COpt::eNoArg) {
                optI->second = "";
            } else { //should have parameter(s)
                for (++argI;  argI < argc  &&
                         (argv[argI][0] != '-'  ||  strlen(argv[argI]) == 1);
                     ++argI) {
                    if (optI->second.AllowsMultiple()) {
                        optI->second += argv[argI];
                    } else {
                        optI->second = argv[argI];
                    }
                }
                if (argI < argc) {//stopped because of '-'
                    --argI; //undo overshoot
                }
                if (optI->second.HasArg() == COpt::eReqArg  &&  !optI->second) {
                    errString += "\t* option '" + string(argv[argI]) +
                        "' is missing required parameter.\n";
                }
            }
        } else { // unattached non-option
            warnString += "\t* unknown argument on command line '" +
                string(argv[argI]) + "'\n";
        }
    }

    string missingReqd;
    bool override = false;
    for (TOpts::const_iterator optI = opts.begin();  optI != opts.end();
         ++optI) {
        if (optI->second.IsRequired()  &&  !(optI->second)) {
            if (!missingReqd.empty()) {
                missingReqd += ", ";
            }
            missingReqd += optI->first;
        }
        if (optI->second.IsOverrideRequired()  &&  (optI->second)) {
            override = true;
        }
    }
    if (!missingReqd.empty()  &&  !override) {
        errString += "\t* missing required options: " + missingReqd + ".\n";
    }

    if (!errString.empty()) {
        cerr << descript
             << "\nUsage:\n"
             << "   " << argv[0] << " (see below for parameters)" << "\n"
             << "Where\n";
        for (TOpts::const_iterator optI = opts.begin();  optI != opts.end();
             ++optI) {
            if (optI->second.GetDesc().empty()) {
                continue;
            }
            cerr << "   " << (!optI->second.IsRequired() ? "[":" ") << "-"
                 << optI->first;
            if (optI->second.HasArg() == COpt::eReqArg) {
                cerr << " <> ";
            } else if (optI->second.HasArg() == COpt::eOptArg) {
                cerr << " [<>] ";
            }
            cerr << (!optI->second.IsRequired() ? "]":"") << " --> "
                 << optI->second.GetDesc();
            if (defaultOpts[optI->first.c_str()]) {
                cerr << " (default " << defaultOpts[optI->first.c_str()] << ")";
            }
            cerr << endl;
        }
        if (!warnString.empty()) {
            cerr << "Warning:\n" << warnString << endl;
        }
        if (!errString.empty()) {
            errString.erase(errString.begin()); //erase leading \t
            throw runtime_error(errString);
        }
    } else if (!warnString.empty()) {
        cerr << "Warning:\n" << warnString << endl;
    }

}

/*---------------------------------------------------------------------------*/

inline void x_ProcessOptions(int argc, char **argv, const char* params,
                      const char* required, TOpts &opts,
                      const char* usage) {
    int c;
    bool err = false;
    bool missingReqd = false;

    while ((c = getopt(argc, argv, params)) != -1) {
        if (c == '?') {
            err = true;
        } else {
            opts[c] = (strchr(params, c)[1] == ':') ? optarg : argv[optind-1];
        }
    }

    for (const char* scan = required;  *scan  &&  !err;  ++scan) {
        if (!opts[*scan]) {
            missingReqd = true;
        }
    }

    if (err  ||  missingReqd) {
        cerr << "Usage:\n"
             << "   " << argv[0] << usage
             << endl;
        if (*required != '\0') { //i.e. we have required elements
            cerr << "NOTE: Parameters '" << required << "' are required."
                 << endl;
        }
        if (err) {
            throw runtime_error("Invalid arguments.");
        } else {
            throw runtime_error("Missing required argument.");
        }
    }
}

/*---------------------------------------------------------------------------*/
// PRE : integer / double
// POST: string representation (this func should really have been in spec)
inline string IntToString(int i) {
    ostringstream oss;
    oss << i << flush;
    return oss.str();
}
inline string DblToString(double d) {
    ostringstream oss;
    oss << d << flush;
    return oss.str();
}

/*---------------------------------------------------------------------------*/
// PRE : open FILE* for reading
// POST: string contains line read until "\n" or eof
inline FILE* getline(FILE* infile, string &s) {
    s.clear();
    
    char buff[100] = "";
    do {
        fgets(buff, 100, infile);
        s += buff;
    } while (*s.rbegin() != '\n'  &&  !feof(infile));

    if (!s.empty()  &&  *s.rbegin() == '\n') {
        s.resize(s.length()-1);
    }

    return infile;
}

/*---------------------------------------------------------------------------*/
// PRE : two values, logged (log[A], log[B])
// POST: the log of the sum (log[A+B])
inline double SumLogged(const double logA, const double logB) {
    if (isinf(logA)  ||  logB > logA + 500) {
        return logB;
    }
    if (isinf(logB)  ||  logA > logB + 500) {
        return logA;
    }
    if (logB > logA) {
        return logA + log(1 + exp(logB - logA));
    } else {
        return logB + log(1 + exp(logA - logB));
    }
}

/*---------------------------------------------------------------------------*/
// PRE : two values, logged (log[A], log[B]); epsilon -- if |logA-logB| < eps,
// consider equal
// POST: the log of the difference (log[A-B])
inline double SubLogged(const double logA, const double logB,
                        const double eps=1e-9) {
    if (isinf(logB)  ||  logA > logB + 500) {
        return logA;
    }
    if (fabs(logA - logB) < eps) {
        return log(0);
    }
    return logA + log(1 - exp(logB - logA));
}

class CLogged {
public:
    CLogged(double d) : m_D(d) {}
    CLogged& operator= (double d) { m_D = log(d); return *this;}
    double exp(void) const { return ::exp(m_D); }

    CLogged& operator*= (const CLogged &l) {
        m_D += l.m_D;
        return *this;
    }
    CLogged& operator/= (const CLogged &l) {
        m_D -= l.m_D;
        return *this;
    }
    CLogged& operator+= (const CLogged &l) {
        m_D = SumLogged(m_D, l.m_D);
        return *this;
    }
    CLogged& operator-= (const CLogged &l) {
        m_D = SubLogged(m_D, l.m_D);
        return *this;
    }
private:
    double m_D;
};

#endif /*ifndef UTIL_HPP*/
